%{

/*#undef yywrap*/

#include <stdio.h>
#include <assert.h>

#include "cobpp.h"
#include "parser.h"	/* definitions from YACC */
/*#include "y.tab.h"*/	/* definitions from YACC */

/*#define DEBUG_COBPP_SCANNER 		1*/

#define MAX_COPY_DEPTH 		10
#define MAX_CPYNAME_LEN 	65
#define MAXLINEBUF         	256
#define MAXBUF         		256

#define FIXFORM_6      		6
#define FIXFORM_7      		7
#define FIXFORM_71     		71
#define FIXFORM_72     		72
#define FIXFORM_73     		73

#define lexchar input

static struct fstack {
	YY_BUFFER_STATE buffer;
	/*int line; */
	struct copy_symbols *last_symbol;
} file_stack[10];
 
static struct copy_symbols {
	struct copy_symbols *next;
	char *value;
	char *replacement;
} *copy_symbols=NULL;
struct copy_symbols *append_copy_symbols=NULL;

char *include_cpybook;
char *include_cpylib;

char include_full_filename[MAXBUF];
char toktext[MAXBUF];

extern FILE *yylist;

struct reserved_sym {
	char *name;
	unsigned int token;
} reserved_symbols[] = {
	{"COPY",TOK_COPY},
	{"OFF",TOK_OFF},
	{"OF",TOK_OF},
	{"IN",TOK_IN},
	{"BY",TOK_BY},
	{"REPLACE",TOK_REPLACE},
	{"REPLACING",TOK_REPLACING},
	{NULL,0}
};

int i;
unsigned int ret_token;
int copy_state_sw = 0;
int copy_start_sw = 0;
int replace_state_sw = 0;
int replace_start_sw = 0;
int previous_state = 0;
int initial_state = 0;
int lf_add = 0;
int quote_sw = 0;
unsigned long columnLen;
unsigned long columnCount = 0;
unsigned long lineCount = 1;
char lineout[80];

void lineExpand(void);
void tabExpand(void);
void lexError( char * msg );
char *savename(const char *s, int x1, int x2);
void debug_rtn (const char *s, unsigned long Ccnt, unsigned long Lcnt, int Len, const char *s1);

void copy_switch(void);
void copy_do_replacements(char *buffer);
void copy_do_replacements_bk(char *buffer);
char *copy_strip_spaces(char *s);
int lex_fgets(char *buf, int maxsize);


#define YY_INPUT(buf,result,max_size) \
	{ \
		result = lex_fgets( buf, max_size ); \
	}

%}

eol	     [\n]
tab	     [\t]
blank        [ ]
Quote	     ['"]
NoQuoteEol   [^'"\n]
DQuote	     ["]
SQuote	     [']
ws	     [ \n]
Period	     \.
Alpha        [A-Za-z]
AlphaNum     [A-Za-z0-9\-]
Numeric      [0-9]
AnyNoDQuote  [^"\t\n]
AnyNoSQuote  [^'\t\n]
EqSign	     [=]
EqSign2	     "=="
NoEqSign     [^\n=]

%x INITAL FREE FIXED FIXED_7 FIXED_A FIXED_LC COMMENT COMMENT_FREE FIXED_ALPHA_LIT

%%

%{
      struct copy_symbols *tmp;
%}
<*>{tab}	{ tabExpand();  }
<FREE>{eol}	{ 
                 debug_rtn ("FREE - {eol} ", columnCount, lineCount, yyleng, "<NL>");
                 lineCount++;
                 columnCount=0; 
		 ECHO;
		 if (yylist != NULL) { 
                    fprintf(yylist, "\n%06d ", lineCount*10); 
                 }
		 if (copy_start_sw != 0) {
                     if (globalEnvPtr->debugFlag != 0) {
                        for (tmp = copy_symbols, i=0; tmp!=NULL; tmp=tmp->next, i++) {
                            fprintf(stderr, "lextrace: <FREE>{eol}  copy_symbols(%d) value=%s replacement=%s\n",
        	            i,
        	            tmp->value,
        	            tmp->replacement 
        	            );
                        }
                     }
                     copy_switch();
                 } 
}

<FREE>^"/"       |
<FREE>^"$"       |       
<FREE>^"D"       |
<FREE>^"*"       { 
            	 columnCount = columnCount + yyleng;
                 debug_rtn ("FREE - *|/|$", columnCount, lineCount, yyleng, yytext);
 		 if (yylist != NULL) { 
		    fprintf(yylist, "%s", yytext);
		 } 
		 /* remove lines */  
		BEGIN(COMMENT_FREE);
}

<FREE>{blank} 	{
                debug_rtn ("FREE - blank", columnCount, lineCount, yyleng, yytext);
	        columnCount++; 
 		if (yylist != NULL) { 
		   fprintf(yylist, "%s", yytext);
		} 
                ECHO; 
}

<FREE>{Numeric}{1,18} {
        debug_rtn ("FREE - {Numeric}", columnCount, lineCount, yyleng, yytext);
	columnCount = columnCount + yyleng;
 	if (yylist != NULL) { 
	   fprintf(yylist, "%s", yytext);
	} 
        if (copy_state_sw != 0) {
           strncpy(toktext, yytext, MAXBUF);   
	   yylval.str = savename(toktext, 0, 0);
 	   return(TOK_LITERAL);
	}
	else {
           ECHO; 
	}

}
<FREE>{EqSign2}{NoEqSign}+{EqSign2}  {
        debug_rtn ("FREE - {EqSign2}{NoEqSign}+{EqSign2}", columnCount, lineCount, yyleng, yytext);
	columnCount = columnCount + yyleng;

 	if (yylist != NULL) { 
	   fprintf(yylist, "%s", yytext);
	} 
        if (copy_state_sw != 0) {
           if (globalEnvPtr->debugFlag != 0) {
               fprintf(stderr, 
            	   "lextrace: TOK_PSEUDO_TEXT :%s: copy_state_sw=%d\n", 
            	   yytext, copy_state_sw);
           }
           strncpy(toktext, yytext, MAXBUF);   
	   yylval.str = savename(toktext, 2, 2);
 	   return(TOK_PSEUDO_TEXT);
	}
	else {
           ECHO; 
	}
}
<FREE>\"{AnyNoDQuote}+\" {
        debug_rtn ("FREE - {DQuote}{AnyNoDQuote}+{DQuote}", columnCount, lineCount, yyleng, yytext);
	columnCount = columnCount + yyleng;
 	if (yylist != NULL) { 
	   fprintf(yylist, "%s", yytext);
	} 
        if (copy_state_sw != 0) {
           if (globalEnvPtr->debugFlag != 0) {
               fprintf(stderr, 
            	   "lextrace: TOK_LITERAL :%s: copy_state_sw=%d\n", 
            	   yytext, copy_state_sw);
           }
           strncpy(toktext, yytext, MAXBUF);   
	   /* yylval.str = savename(toktext, 1, 1); */
	   yylval.str = savename(toktext, 0, 0);
 	   return(TOK_LITERAL);
	}
	else {
           ECHO; 
	}

}
<FREE>\'{AnyNoSQuote}+\' {
        debug_rtn ("FREE - {SQuote}{AnyNoSQuote}+{SQuote}", columnCount, lineCount, yyleng, yytext);
	columnCount = columnCount + yyleng;

 	if (yylist != NULL) { 
	   fprintf(yylist, "%s", yytext);
	} 

        if (copy_state_sw != 0) {
           if (globalEnvPtr->debugFlag != 0) {
               fprintf(stderr, 
            	   "lextrace: TOK_LITERAL :%s: copy_state_sw=%d\n", 
            	   yytext, copy_state_sw);
           }
           strncpy(toktext, yytext, MAXBUF);   
	   /*yylval.str = savename(toktext, 1, 1);*/
	   yylval.str = savename(toktext, 0, 0);
 	   return(TOK_LITERAL);
	}
	else {
           ECHO; 
	}
}
<FREE>{Period}/[ \n]  {
         debug_rtn ("FREE - {Period}", columnCount, lineCount, yyleng, yytext);
	 columnCount++;
 	 if (yylist != NULL) { 
	    fprintf(yylist, "%s", yytext);
  	 } 
	 if (copy_state_sw != 0) {
	    return(TOK_PERIOD);
	 }
	 else {
           ECHO; 
	}
}
<FREE>[cC][oO][pP][yY] {
        debug_rtn ("FREE - [cC][oO][pP][yY] ", columnCount, lineCount, yyleng, yytext);
	columnCount = columnCount + yyleng;

 	if (yylist != NULL) { 
	   fprintf(yylist, "%s", yytext);
	} 

        copy_state_sw++;
        if (globalEnvPtr->debugFlag != 0) {
           fprintf(stderr, 
                 "lextrace: TOK_COPY=%d, copy_state_sw=%d;\n", 
                 TOK_COPY, copy_state_sw);
        }
	return TOK_COPY;
}
<FREE>[oO][Ff] {
        debug_rtn ("FREE - [oO][Ff] ", columnCount, lineCount, yyleng, yytext);
	columnCount = columnCount + yyleng;

 	if (yylist != NULL) { 
	   fprintf(yylist, "%s", yytext);
	} 
        if (globalEnvPtr->debugFlag != 0) {
           fprintf(stderr, 
                 "lextrace: TOK_OF=%d, copy_state_sw=%d;\n", 
                 TOK_OF, copy_state_sw);
        }
        if (copy_state_sw != 0) {
	    return TOK_OF;
        }
        else {
            ECHO;
        }
}
<FREE>[oO][Ff][Ff] {
        debug_rtn ("FREE - [oO][Ff][Ff] ", columnCount, lineCount, yyleng, yytext);
	columnCount = columnCount + yyleng;

 	if (yylist != NULL) { 
	   fprintf(yylist, "%s", yytext);
	} 
        if (globalEnvPtr->debugFlag != 0) {
           fprintf(stderr, 
                 "lextrace: TOK_OFF=%d, copy_state_sw=%d;\n", 
                 TOK_OFF, copy_state_sw);
        }
	return TOK_OFF;
}

<FREE>[Ii][nN] {
        debug_rtn ("FREE - [Ii][nN] ", columnCount, lineCount, yyleng, yytext);
	columnCount = columnCount + yyleng;

 	if (yylist != NULL) { 
	   fprintf(yylist, "%s", yytext);
	} 
        if (globalEnvPtr->debugFlag != 0) {
           fprintf(stderr, 
                 "lextrace: TOK_IN=%d, copy_state_sw=%d;\n", 
                 TOK_IN, copy_state_sw);
        }
        if (copy_state_sw != 0) {
	    return TOK_IN;
        }
        else {
            ECHO;
        }
}

<FREE>[Bb][yY] {
        debug_rtn ("FREE - [Bb][yY] ", columnCount, lineCount, yyleng, yytext);
	columnCount = columnCount + yyleng;

 	if (yylist != NULL) { 
	   fprintf(yylist, "%s", yytext);
	} 
        if (globalEnvPtr->debugFlag != 0) {
           fprintf(stderr, 
                 "lextrace: TOK_BY=%d, copy_state_sw=%d;\n", 
                 TOK_BY, copy_state_sw);
        }
        if (copy_state_sw != 0) {
	    return TOK_BY;
        }
        else {
            ECHO;
        }
}
<FREE>[Rr][Ee][Pp][Ll][Aa][Cc][Ee] {
        debug_rtn ("FREE - [Rr][Ee][Pp][Ll][Aa][Cc][Ee] ", columnCount, lineCount, yyleng, yytext);
	columnCount = columnCount + yyleng;

 	if (yylist != NULL) { 
	   fprintf(yylist, "%s", yytext);
	} 
        if (globalEnvPtr->debugFlag != 0) {
           fprintf(stderr, 
                 "lextrace: TOK_REPLACE=%d, copy_state_sw=%d;\n", 
                 TOK_REPLACE, copy_state_sw);
        }
	return TOK_REPLACE;
}

<FREE>[Rr][Ee][Pp][Ll][Aa][Cc][iI][nN][Gg] {
        debug_rtn ("FREE - [Rr][Ee][Pp][Ll][Aa][Cc][iI][nN][Gg] ", columnCount, lineCount, yyleng, yytext);
	columnCount = columnCount + yyleng;

 	if (yylist != NULL) { 
	   fprintf(yylist, "%s", yytext);
	} 
        if (globalEnvPtr->debugFlag != 0) {
           fprintf(stderr, 
                 "lextrace: TOK_REPLACING=%d, copy_state_sw=%d;\n", 
                 TOK_REPLACING, copy_state_sw);
        }
	if (copy_state_sw != 0) {
	    return TOK_REPLACING;
        }
        else {
            ECHO;
        }
}
<FREE>{Alpha}+{AlphaNum}* {
                debug_rtn ("FREE - {Alpha}+{AlphaNum}*", columnCount, lineCount, yyleng, yytext); 
	        columnCount = columnCount + yyleng;

 		if (yylist != NULL) { 
		   fprintf(yylist, "%s", yytext);
		} 

	        if (copy_state_sw != 0) {
                   if (globalEnvPtr->debugFlag != 0) {
                      fprintf(stderr, 
                     	  "lextrace: TOK_IDENTIFIER :%s: copy_state_sw=%d\n", 
                     	  yytext, copy_state_sw);
                   }
                   strncpy(toktext, yytext, MAXBUF);   
	           yylval.str = savename(toktext, 0, 0);
 	           return(TOK_IDENTIFIER);
	        }
	        else {
                   ECHO;
	        }
} 
<FREE>. 	{ 
                debug_rtn ("FREE - .", columnCount, lineCount, yyleng, yytext);
		columnCount++;

 		if (yylist != NULL) { 
		   fprintf(yylist, "%s", yytext);
		} 
	        ECHO;
}	
<COMMENT_FREE>. { 
                 debug_rtn ("COMMENT_FREE - . ", columnCount, lineCount, yyleng, yytext);
		 columnCount++;
 		 if (yylist != NULL) { 
		    fprintf(yylist, "%s", yytext);
		 }
}
<COMMENT_FREE>{eol} { 
                 debug_rtn ("COMMENT_FREE - {eol} ", columnCount, lineCount, yyleng, "<NL>");
                 lineCount++;
                 columnCount=0; 
		 ECHO;
		 if (yylist != NULL) { 
                    fprintf(yylist, "\n%06d ", lineCount*10); 
                 }
		 BEGIN(FREE);
}

<FIXED_LC>{eol}.{6}"-"{blank}{4,58}{Quote}	 { 
                 debug_rtn ("FIXED_LC - {eol}.{6}-{blank}{4}{Quote}", columnCount, lineCount, yyleng, yytext);
		 /* Continuation line remove */ 
		 lineCount++;
		 columnCount = yyleng - 1; 
 		 if (yylist != NULL) { 
		    fprintf(yylist, "\n%06d %s", lineCount*10, yytext+1);
		 } 
		 lf_add++; 
		 BEGIN(FIXED_ALPHA_LIT);
}
					
<FIXED_LC>{eol} { 
                debug_rtn ("FIXED_LC - {eol} ", columnCount, lineCount, yyleng, "<NL>");
		lineCount++;
		columnCount=0; 
		ECHO;

		for (i=0; i<lf_add; i++) {
                    fprintf(yyout, "\n"); 
		}
		lf_add=0;
 		if (yylist != NULL) { 
		    fprintf(yylist, "%s", yytext);
		} 

		if (copy_start_sw != 0) {
                    copy_switch();
                } 
		
                BEGIN(FIXED);
}

<COMMENT>{eol} { 
                 debug_rtn ("COMMENT - {eol} ", columnCount, lineCount, yyleng, "<NL>");
		 lineCount++;
		 columnCount=0; 
		 ECHO; 
 		 if (yylist != NULL) { 
		    fprintf(yylist, "%s", yytext);
		 } 
		 BEGIN(FIXED); 
}		

<FIXED_LC>. { 
                 debug_rtn ("FIXED_LC - .", columnCount, lineCount, yyleng, yytext);
		 columnCount++;
		 
		 if ( columnCount > 80 ) { 
 		      fprintf(stderr,
			 "%s:%lu:(2) Invaild character (%c) in column %lu.\n",
			 globalEnvPtr->ifname,
			 lineCount,
			 yytext[0],
			 columnCount
			 );
		 }
                 
 		 if (yylist != NULL) { 
		    fprintf(yylist, "%s", yytext);
		 } 
		 
}

<FIXED>{eol} { 
                 debug_rtn ("FIXED - {eol}", columnCount, lineCount, yyleng, "<NL>");
		 lineCount++; 
		 columnCount=0; 
		 ECHO;
		 for (i=0; i<lf_add; i++) {
                    fprintf(yyout, "\n"); 
		 }		  
 		 if (yylist != NULL) { 
		    fprintf(yylist, "\n");
		 } 
		 lf_add=0; 
		 BEGIN(FIXED);
}
<FIXED>^.{6}"-"{blank}{4,58}{Quote}	 { 
		 columnCount = columnCount + yyleng; 
                 debug_rtn ("FIXED - ^.{6}-{blank}{4}{Quote}", columnCount, lineCount, yyleng, yytext);
		 /* Continuation line remove */ 
 		 if (yylist != NULL) { 
		    fprintf(yylist, "%06d %s", lineCount*10, yytext);
		 }
		 lf_add++; 
		 BEGIN(FIXED_ALPHA_LIT);
}
					
<FIXED>^.{6} {
                 debug_rtn ("FIXED - ^.{6}", columnCount, lineCount, yyleng, yytext);
		 columnCount+=6; 
 		 if (yylist != NULL) { 
		    fprintf(yylist, "%06d %s", lineCount*10, yytext);
		 } 
		 BEGIN(FIXED_7); 
}		
<FIXED_7>"*"	|
<FIXED_7>"/" 	|
<FIXED_7>"$" 	{ 
                 debug_rtn ("FIXED_7 - *|/|$", columnCount, lineCount, yyleng, yytext);
		 columnCount++;
 		 if (yylist != NULL) { 
		    fprintf(yylist, "%s", yytext);
		 } 
		 /* comment line remove */ 
		 BEGIN(COMMENT);
}
	
<FIXED_7>"D" { 
                 debug_rtn ("FIXED_7 - D", columnCount, lineCount, yyleng, yytext);
		 columnCount++;
 		 if (yylist != NULL) { 
		    fprintf(yylist, "%s", yytext);
		 } 
		 BEGIN(COMMENT);  
}

<FIXED_7>[^ \n]	{ 
                 debug_rtn ("FIXED_7 - [^ nl]", columnCount, lineCount, yyleng, yytext);
		 globalEnvPtr->errFlag++;
		 fprintf(stderr,
		    	  "%s:%lu:(3) Invaild character (%c) in column %lu.\n",
			  globalEnvPtr->ifname,
			  lineCount,
			  yytext[0],
			  columnCount
			 );
 		 if (yylist != NULL) { 
		    fprintf(yylist, "%s", yytext);
		 } 
		 BEGIN(COMMENT);

}

<FIXED_7>{eol}	{ 
                 debug_rtn ("FIXED_7 - {eol}", columnCount, lineCount, yyleng, "<NL>");
		 lineCount++;
		 columnCount=0;
		 ECHO;
 		 if (yylist != NULL) { 
		    fprintf(yylist, "\n");
		 } 
		 for (i=0; i<lf_add; i++) {
                    fprintf(yyout, "\n"); 
		 }
		 lf_add=0;
		 if (copy_start_sw != 0) {
                    copy_switch();
                 } 
		 BEGIN(FIXED); 
}

<FIXED_7>{blank} { 
                 debug_rtn ("FIXED_7 - blank", columnCount, lineCount, yyleng, yytext);
		 columnCount++;
 		 if (yylist != NULL) { 
		    fprintf(yylist, "%s", yytext);
		 } 
		 ECHO;
                 if (previous_state == FIXED_ALPHA_LIT) {
                    BEGIN(FIXED_ALPHA_LIT);
                 }
                 else {
                    BEGIN(FIXED_A);
                 }
}
	
<FIXED_A>{blank}   {
                debug_rtn ("FIXED_A - blank", columnCount, lineCount, yyleng, yytext);
	        columnCount++; 
 		if (yylist != NULL) { 
		   fprintf(yylist, "%s", yytext);
		} 
	        ECHO;
	        if (columnCount < 72) {
		   BEGIN(FIXED_A); 
	        }
	        else {
	           BEGIN(FIXED_LC);
	        }
}

<FIXED_A>{Numeric}{1,18} {
        debug_rtn ("FIXED_A - {Numeric}", columnCount, lineCount, yyleng, yytext);
	columnCount = columnCount + yyleng;
 	if (yylist != NULL) { 
	   fprintf(yylist, "%s", yytext);
	} 
        if (copy_state_sw != 0) {
           strncpy(toktext, yytext, MAXBUF);   
	   yylval.str = savename(toktext, 0, 0);
 	   return(TOK_LITERAL);
	}
	else {
           ECHO; 
	}

}

<FIXED_A>{EqSign2}{NoEqSign}+{EqSign2}  {
        debug_rtn ("FIXED_A - {EqSign2}{NoEqSign}+{EqSign2}", columnCount, lineCount, yyleng, yytext);
	columnCount = columnCount + yyleng;

 	if (yylist != NULL) { 
	   fprintf(yylist, "%s", yytext);
	} 
        if (copy_state_sw != 0) {
           if (globalEnvPtr->debugFlag != 0) {
               fprintf(stderr, 
            	   "lextrace: TOK_PSEUDO_TEXT :%s: copy_state_sw=%d\n", 
            	   yytext, copy_state_sw);
           }
           strncpy(toktext, yytext, MAXBUF);   
	   yylval.str = savename(toktext, 2, 2);
 	   return(TOK_PSEUDO_TEXT);
	}
	else {
           ECHO; 
	}
}

<FIXED_ALPHA_LIT>\" {
        debug_rtn ("FIXED_ALPHA_LIT - {DQuote}", columnCount, lineCount, yyleng, yytext);
	columnCount++;
 	if (yylist != NULL) { 
	   fprintf(yylist, "%s", yytext);
	} 
        if (copy_state_sw != 0) {
           if (globalEnvPtr->debugFlag != 0) {
               fprintf(stderr, 
            	   "lextrace: TOK_LITERAL :%s: copy_state_sw=%d\n", 
            	   yytext, copy_state_sw);
           }
           strcat(toktext, yytext);   
	   /* yylval.str = savename(toktext, 1, 1); */
	   yylval.str = savename(toktext, 0, 0);
           previous_state = 0;
           quote_sw = 0;
           BEGIN(FIXED_A);
 	   return(TOK_LITERAL);
	}
	else {
           ECHO; 
           previous_state = 0;
           quote_sw = 0;
           BEGIN(FIXED_A);
	}
}

<FIXED_ALPHA_LIT>\' {
        debug_rtn ("FIXED_ALPHA_LIT - {SQuote}", columnCount, lineCount, yyleng, yytext);
	columnCount++;

 	if (yylist != NULL) { 
	   fprintf(yylist, "%s", yytext);
	} 

        if (copy_state_sw != 0) {
           if (globalEnvPtr->debugFlag != 0) {
               fprintf(stderr, 
            	   "lextrace: TOK_LITERAL :%s: copy_state_sw=%d\n", 
            	   yytext, copy_state_sw);
           }
           strcat(toktext, yytext);
	   if (quote_sw == 1) {
 	       /* yylval.str = savename(toktext, 1, 1); */
	       yylval.str = savename(toktext, 0, 0);
               previous_state = 0;
	       quote_sw = 0;
	       BEGIN(FIXED_A);
 	       return(TOK_LITERAL);
           }
	}
	else {
           ECHO; 
	   if (quote_sw == 1) {
              previous_state = 0;
              quote_sw = 0;
              BEGIN(FIXED_A);
           }
	}
}

<FIXED_ALPHA_LIT>{NoQuoteEol} {
        debug_rtn ("FIXED_ALPHA_LIT - {NoQuoteEol}", columnCount, lineCount, yyleng, yytext);
	columnCount++;

 	if (yylist != NULL) { 
	   fprintf(yylist, "%s", yytext);
	} 
        
        if (copy_state_sw != 0) {
           strcat(toktext, yytext);
	}
	else {
           ECHO; 
	}

        if (columnCount == FIXFORM_72) {
           previous_state = FIXED_ALPHA_LIT;
           BEGIN(FIXED_LC);
        }
        
}

<FIXED_ALPHA_LIT>{eol} {
        debug_rtn ("FIXED_ALPHA_LIT - {eol}", columnCount, lineCount, yyleng, yytext);
	columnCount=0;
        lineCount++; 

 	if (yylist != NULL) { 
	   fprintf(yylist, "\n");
	} 

        if (copy_state_sw != 0) {
           if (globalEnvPtr->debugFlag != 0) {
               fprintf(stderr, 
            	   "lextrace: TOK_LITERAL :%s: copy_state_sw=%d\n", 
            	   yytext, copy_state_sw);
           }
	}
	else {
           BEGIN(FIXED);
	}
}

<FIXED_A>\" {
        debug_rtn ("FIXED_A - {DQuote}", columnCount, lineCount, yyleng, yytext);
	columnCount++;
	quote_sw = 2;
 	if (yylist != NULL) { 
	   fprintf(yylist, "%s", yytext);
	} 
        if (copy_state_sw != 0) {
           strncpy(toktext, yytext, MAXBUF);   
	}
	else {
           ECHO; 
	}
        previous_state = 0;
        BEGIN(FIXED_ALPHA_LIT);
}

<FIXED_A>\' {
        debug_rtn ("FIXED_A - {SQuote}", columnCount, lineCount, yyleng, yytext);
	columnCount++;
	quote_sw = 1;
 	if (yylist != NULL) { 
	   fprintf(yylist, "%s", yytext);
	} 
        if (copy_state_sw != 0) {
           strncpy(toktext, yytext, MAXBUF);   
	}
	else {
           ECHO; 
	}
        previous_state = 0;
        BEGIN(FIXED_ALPHA_LIT);
}

<FIXED_A>{Period}/[ \n]  {
         debug_rtn ("FIXED_A - {Period}", columnCount, lineCount, yyleng, yytext);
	 columnCount++;
 	 if (yylist != NULL) { 
	    fprintf(yylist, "%s", yytext);
  	 } 
	 if (copy_state_sw != 0) {
	    return(TOK_PERIOD);
	 }
	 else {
           ECHO; 
	}
}
           
<FIXED_A>[cC][oO][pP][yY] {
        debug_rtn ("FIXED_A - [cC][oO][pP][yY] ", columnCount, lineCount, yyleng, yytext);
	columnCount = columnCount + yyleng;

 	if (yylist != NULL) { 
	   fprintf(yylist, "%s", yytext);
	} 

        copy_state_sw++;
        if (globalEnvPtr->debugFlag != 0) {
           fprintf(stderr, 
                 "lextrace: TOK_COPY=%d, copy_state_sw=%d;\n", 
                 TOK_COPY, copy_state_sw);
        }
	return TOK_COPY;
}
<FIXED_A>[oO][Ff] {
        debug_rtn ("FIXED_A - [oO][Ff] ", columnCount, lineCount, yyleng, yytext);
	columnCount = columnCount + yyleng;

 	if (yylist != NULL) { 
	   fprintf(yylist, "%s", yytext);
	} 
        if (globalEnvPtr->debugFlag != 0) {
           fprintf(stderr, 
                 "lextrace: TOK_OF=%d, copy_state_sw=%d;\n", 
                 TOK_OF, copy_state_sw);
        }
        if (copy_state_sw != 0) {
	    return TOK_OF;
        }
        else {
            ECHO;
        }
}

<FIXED_A>[oO][Ff][Ff] {
        debug_rtn ("FIXED_A - [oO][Ff][Ff] ", columnCount, lineCount, yyleng, yytext);
	columnCount = columnCount + yyleng;

 	if (yylist != NULL) { 
	   fprintf(yylist, "%s", yytext);
	} 
        if (globalEnvPtr->debugFlag != 0) {
           fprintf(stderr, 
                 "lextrace: TOK_OFF=%d, copy_state_sw=%d;\n", 
                 TOK_OFF, copy_state_sw);
        }
	return TOK_OFF;
}

<FIXED_A>[Ii][nN] {
        debug_rtn ("FIXED_A - [Ii][nN] ", columnCount, lineCount, yyleng, yytext);
	columnCount = columnCount + yyleng;

 	if (yylist != NULL) { 
	   fprintf(yylist, "%s", yytext);
	} 
        if (globalEnvPtr->debugFlag != 0) {
           fprintf(stderr, 
                 "lextrace: TOK_IN=%d, copy_state_sw=%d;\n", 
                 TOK_IN, copy_state_sw);
        }
        if (copy_state_sw != 0) {
	    return TOK_IN;
        }
        else {
            ECHO;
        }
}

<FIXED_A>[Bb][yY] {
        debug_rtn ("FIXED_A - [Bb][yY] ", columnCount, lineCount, yyleng, yytext);
	columnCount = columnCount + yyleng;

 	if (yylist != NULL) { 
	   fprintf(yylist, "%s", yytext);
	} 
        if (globalEnvPtr->debugFlag != 0) {
           fprintf(stderr, 
                 "lextrace: TOK_BY=%d, copy_state_sw=%d;\n", 
                 TOK_BY, copy_state_sw);
        }
        if (copy_state_sw != 0) {
	    return TOK_BY;
        }
        else {
            ECHO;
        }
}

<FIXED_A>[Rr][Ee][Pp][Ll][Aa][Cc][Ee] {
        debug_rtn ("FIXED_A - [Rr][Ee][Pp][Ll][Aa][Cc][Ee] ", columnCount, lineCount, yyleng, yytext);
	columnCount = columnCount + yyleng;

 	if (yylist != NULL) { 
	   fprintf(yylist, "%s", yytext);
	} 
        if (globalEnvPtr->debugFlag != 0) {
           fprintf(stderr, 
                 "lextrace: TOK_REPLACE=%d, copy_state_sw=%d;\n", 
                 TOK_REPLACE, copy_state_sw);
        }
	return TOK_REPLACE;
}

<FIXED_A>[Rr][Ee][Pp][Ll][Aa][Cc][iI][nN][Gg] {
        debug_rtn ("FIXED_A - [Rr][Ee][Pp][Ll][Aa][Cc][iI][nN][Gg] ", columnCount, lineCount, yyleng, yytext);
	columnCount = columnCount + yyleng;

 	if (yylist != NULL) { 
	   fprintf(yylist, "%s", yytext);
	} 
        if (globalEnvPtr->debugFlag != 0) {
           fprintf(stderr, 
                 "lextrace: TOK_REPLACING=%d, copy_state_sw=%d;\n", 
                 TOK_REPLACING, copy_state_sw);
        }
	if (copy_state_sw != 0) {
	    return TOK_REPLACING;
        }
        else {
            ECHO;
        }
}
           
<FIXED_A>{Alpha}+{AlphaNum}* {
                debug_rtn ("FIXED_A - {Alpha}+{AlphaNum}*", columnCount, lineCount, yyleng, yytext); 
	        columnCount = columnCount + yyleng;

 		if (yylist != NULL) { 
		   fprintf(yylist, "%s", yytext);
		} 

	        if (copy_state_sw != 0) {
                   if (globalEnvPtr->debugFlag != 0) {
                      fprintf(stderr, 
                     	  "lextrace: TOK_IDENTIFIER :%s: copy_state_sw=%d\n", 
                     	  yytext, copy_state_sw);
                   }
                   strncpy(toktext, yytext, MAXBUF);   
	           yylval.str = savename(toktext, 0, 0);
 	           return(TOK_IDENTIFIER);
	        }
	        else {
                   ECHO;
	        }
	        BEGIN(FIXED_A);
} 
            
<FIXED_A>.      { 
                debug_rtn ("FIXED_A - .", columnCount, lineCount, yyleng, yytext);
                columnCount++;
 		if (yylist != NULL) { 
		   fprintf(yylist, "%s", yytext);
		} 
	        ECHO;
	        if (columnCount < 71) {
		   BEGIN(FIXED_A); 
	        }
	        else {
	           BEGIN(FIXED_LC);
	        }
}	

<COMMENT>. { 
                 debug_rtn ("COMMENT - . ", columnCount, lineCount, yyleng, yytext);
            	 columnCount = columnCount + yyleng; 
		 /*
		 if ( columnCount > 80 ) { 
 		      fprintf(stderr,
			 "%s:%lu:(2) Invaild character (%c) in column %lu.\n",
			 globalEnvPtr->ifname,
			 lineCount,
			 yytext[0],
			 columnCount
			 );
		 }
		 */
 		 if (yylist != NULL) { 
		    fprintf(yylist, "%s", yytext);
		 } 
		 if ( columnCount > 71 ) {
 		    /*ECHO;*/
		    BEGIN(FIXED_LC);
		 }
		 

}


<FIXED_A>{eol} { 
                 debug_rtn ("FIXED_A - {eol} ", columnCount, lineCount, yyleng, "<NL>");
		 lineCount++; 
		 columnCount=0; 
		 ECHO;
		 if (yylist != NULL) { 
		   fprintf(yylist,"\n");
		 } 
		 for (i=0; i<lf_add; i++) {
                    fprintf(yyout, "\n"); 
		 }
		 lf_add=0;
		 if (copy_start_sw != 0) {
                     if (globalEnvPtr->debugFlag != 0) {
                        for (tmp = copy_symbols, i=0; tmp!=NULL; tmp=tmp->next, i++) {
                            fprintf(stderr, "lextrace: <FIXED_A>{eol}  copy_symbols(%d) value=%s replacement=%s\n",
        	            i,
        	            tmp->value,
        	            tmp->replacement 
        	            );
                        }
                     }
                     copy_switch();
                 } 
		 BEGIN(FIXED);

}

<<EOF>> {
	int tmp;
	struct copy_symbols *cs,*cs1;
	if (fsp == 0) {
	        if (globalEnvPtr->debugFlag != 0) {
		//fprintf(stderr,"scanner receiving <<EOF>>\n");
                    fprintf(stderr, "lextrace: receiving <<EOF>> for fsp = %d:\n",
    	            fsp
    	            );
    	        }
		yyterminate();
	}
	else {

// 		tmp = lineno;
// 		yy_delete_buffer( YY_CURRENT_BUFFER );
// 		yy_switch_to_buffer( file_stack[fsp].buffer );
// 		lineno = file_stack[ fsp  ].line;
		/* free all replacements in this dead context */
	        if (globalEnvPtr->debugFlag != 0) {
                    fprintf(stderr, "lextrace: begin delete fsp = %d:\n",
    	            fsp
    	            );
    	        }
		cs1 = file_stack[fsp].last_symbol;
		cs = copy_symbols;
		if (cs == cs1)
			copy_symbols = NULL;
		while (cs) {
			if (cs->next == cs1)
				break;
			cs = cs->next;
		}
		if (cs)
			cs->next = NULL; 
		if (cs1) {
			while (cs1) {
                                if (globalEnvPtr->debugFlag != 0) {
                                    fprintf(stderr, "lextrace: delete level=%d, value=%s, replacement=%s:\n",
    	                            fsp,
    	                            cs1->value,
    	                            cs1->replacement
    	                            );
                                }
				free( cs1->value );
				free( cs1->replacement );
				cs = cs1->next;
				cs1->next = NULL;
				free( cs1 );
				cs1 = cs;
				cs = NULL;
			}
		}
		yy_delete_buffer( YY_CURRENT_BUFFER );
		fsp--;
		yy_switch_to_buffer( file_stack[fsp].buffer );

	}
}

%%


/*
 * print error on tab.
 */
void lineExpand(void) {
  int nlines, j;
  char *s = yytext, delm = '0';
  
  for(j=0; j<columnLen; j++) {
     if( (yytext[j] == '\'') || (yytext[j] == '"') ) {
        delm = yytext[j];
        j = columnLen + 3;
     }
  }
  if(delm == '0') {
    fprintf(stderr, "Error: quote/double expected in litteral, not found");
    return;
  }
  strncpy(lineout, s, 66);
  lineout[66] = '\0';
  printf("%s", lineout);
  columnCount = 66;
  s = yytext + 66;
  nlines = ((columnLen - 66) / 60) + 1;

  for(j=1; j<nlines; j++) {
     sprintf(lineout, "-    %c", delm);
     strncat(lineout, s, 60);
     lineout[66] = '\0';
     s = s + 60;
     columnCount =  columnCount + 66;
     lineCount++;
     printf("\n%06d%s", lineCount*10, lineout); 
  }

  if( (j = strlen(s)) > 0) {
    sprintf(lineout, "-    %c", delm);
    strncat(lineout, s, j);
    lineout[6 + j] = '\0';
    lineCount++;
    printf("\n%06d%s", lineCount*10, lineout); 
  }
  		    
/*
    fprintf(stderr,
           "%s: Character (%c) in column %d, line %lu.\n",
	    globalEnvPtr->ifname,
	    yytext[0],
            columnCount,
            lineCount);
*/
}

/* Expand tabs if tabs number defined */
void tabExpand(void) {
  int i, j;
  
  if (globalEnvPtr->tab2space > 0) {
     j = globalEnvPtr->tab2space - (columnCount % globalEnvPtr->tab2space);
     for (i=0; i<j; i++) {
         fprintf(yyout, " "); 
     }
     if (yylist != NULL) { 
        for (i=0; i<j; i++) {
           fprintf(yylist, " ");
        }
     } 
     if (globalEnvPtr->debugFlag != 0) {
        fprintf(stderr, 
           "lex debug: column=%d; line=%d; mod=%d; j=%d;\n", 
           columnCount,
           lineCount,
           (columnCount % globalEnvPtr->tab2space), 
           j 
           ); 
     } 
     columnCount = columnCount + j;
  }
  else {
     fprintf(stderr, "%s:%lu: Error tab in column %lu\n", 
	globalEnvPtr->ifname,
	lineCount, 
	columnCount
	);
      globalEnvPtr->errFlag++;
  }

  return;
}

void lexError( char * msg ) {

	(void)fprintf(stderr,
		"%s:%lu: %s\n"
		,globalEnvPtr->ifname
		,lineCount
		,msg
	);
	globalEnvPtr->errFlag++;
	return;
}
 
int yywrap(void) {
 
/* yywrap will be called to open files from the List of Files in the
 * Global Env struct.
 */
	return 1;
}

char *savename(const char *s, int x1, int x2) {
  char *ap, *s1;
  int len;
  if ((ap = (char *) malloc( strlen(s) + 1 )) != NULL) {
     strcpy(ap, s + x1);
     len = strlen(ap);
     s1 = ap + len - x2;
     *s1 = '\0';
     s1++;
     *s1 = '\0';
#ifdef DEBUG_COBPP_SCANNER
        fprintf(stderr, "lextrace: savename :%s:\n", ap);
#endif
  }
  return(ap);
}

void debug_rtn (const char *s, unsigned long Ccnt, unsigned long Lcnt, int Len, const char *s1) {

#ifdef DEBUG_COBPP_SCANNER
     fprintf(stderr, 
             "debug %s(%lu,%lu): length=%d, :%s:\n", 
             s,
             Lcnt, 
             Ccnt, 
             Len,
             s1
             );   
#endif
}

int lookup_reserved(char *s) {
    
    int i=0;
    
    while (reserved_symbols[i].name != NULL) {
       if (strcasecmp(s, reserved_symbols[i].name) == 0) {
          if (globalEnvPtr->debugFlag != 0) {
              fprintf(stderr, 
                  "lextrace: lookup_reserved=(:%s:, :%s:, %d)\n", 
                  s, reserved_symbols[i].name, reserved_symbols[i].token);
          }
          return reserved_symbols[i].token;
       }
       i++;
    }
    
    return 0;
}

int lex_fgets(char *buf, int maxsize) {
	int result;
	char *s;
	char tbuf[MAXLINEBUF];
	s = fgets(buf, maxsize, yyin);
	result = (s==NULL) ? YY_NULL: strlen(s);
	
	if (globalEnvPtr->debugFlag != 0 ) {
//   	   fprintf(stderr,"lextrace: buf[%d] before: |%s|\n", strlen(buf), buf);
	   if (fsp > 0) {
              fprintf(stderr, "lextrace: fsp = %d:\n",
    	            fsp
    	            );
          }
        }
	
	if ((result) && (fsp > 0)) {
	        strcpy(tbuf, buf);
	        tbuf[ strlen(tbuf) - 1] = CHR_EOS;
                if (globalEnvPtr->debugFlag != 0) {
  		   fprintf(stderr,"lextrace: tbuf[%d] before: |%s|\n", strlen(tbuf), tbuf);
//  		   fprintf(stderr,"lextrace: buf[%d] before: |%s|\n", strlen(buf), buf);
                }	 
		copy_do_replacements( buf ); 
		result = strlen(buf);

	        strcpy(tbuf, buf);
	        tbuf[ strlen(tbuf) - 1] = CHR_EOS;
                if (globalEnvPtr->debugFlag != 0) {
  		   fprintf(stderr,"lextrace:  tbuf[%d] after: |%s|\n", strlen(tbuf), tbuf); 
//  		   fprintf(stderr,"lextrace: buf[%d] after: |%s|\n", strlen(buf), buf); 
                }	 

	} 
	return result;
}

void setup_scanner_state () {

    /*copy_state_sw = 0;
    copy_start_sw = 0;
    replace_state_sw = 0;
    replace_start_sw = 0;*/

    if ( globalEnvPtr->codeFormat == 1 ) {
      /* free == 1, fixed == 0 */
	    if (yylist != NULL) { 
              fprintf(yylist, "%06d ", lineCount*10); 
            }
    	    BEGIN( FREE ) ;
    }
    else {
         if ( globalEnvPtr->codeFormat == 0) {
    	    BEGIN( FIXED ) ;
         }
         else {
    	    assert(1);
         }
    }
}

void copy_switch(void) {

	struct copy_symbols *tmp;

	if (fsp>=MAX_COPY_DEPTH) {
		/* yyerror("Impossible to redirect input!"); */
		fprintf(stderr, "Copy nested too deeply (max=%d)\n", MAX_COPY_DEPTH);
		exit(1);
	}
	file_stack[ fsp ].buffer = YY_CURRENT_BUFFER;
// 	file_stack[ fsp ].line = lineno;
// 	file_stack[ fsp ].copybook = include_cpybook;
// 	file_stack[ fsp ].copylib = include_cpylib;

        if (globalEnvPtr->debugFlag != 0) {
           for (tmp = append_copy_symbols,i=0; tmp!=NULL; tmp=tmp->next, i++) {
               fprintf(stderr, "lextrace: dump2 copy_symbols(%d) value=%s replacement=%s\n",
        	      i,
        	      tmp->value,
        	      tmp->replacement 
        	      );
           }
        }

 	copy_symbols = NULL;

	fsp++;
	file_stack[ fsp ].last_symbol = append_copy_symbols;
        append_copy_symbols = NULL;

        if (globalEnvPtr->debugFlag != 0 ) {
           fprintf(stderr, "lextrace: level=%d, copybook=%s copylib=%s\n", 
             fsp,
	     include_cpybook,
	     include_cpylib
             );
           for (tmp = file_stack[fsp].last_symbol, i=0; tmp!=NULL; tmp=tmp->next, i++) {
               fprintf(stderr, "lextrace: dump3(%d) level=%d, value=%s, replacement=%s\n",
                         i,
                         fsp,
                         tmp->value,
                         tmp->replacement 
                         );
           }
                
        }

        if (find_copybook_file(include_cpybook, include_cpylib) != 0) {

	   fprintf(stderr, "error: copybook file %s not found\n", include_cpybook);

	   exit(1);
        }     
           
	if ((yyin = fopen( include_full_filename, "r" ))==NULL) {
		fprintf(stderr, "Cannot open copy file %s\n", include_full_filename);
		exit(1);
	}

	yy_switch_to_buffer( yy_create_buffer(yyin,YY_BUF_SIZE) );

        copy_start_sw--;

}

void add_copy_replacement(char *orig_value, char *replacement) {

    struct copy_symbols *cs,*tmp;
    int i;

    if (globalEnvPtr->debugFlag != 0 ) {
       fprintf(stderr, "lextrace: add_copy_replacement - %s will be replaced by %s\n",
    	       orig_value,
    	       replacement
    	       );
    }
	cs = malloc(sizeof(struct copy_symbols));
	cs->value = orig_value;
	cs->replacement = replacement;
	cs->next = NULL;
	if (append_copy_symbols==NULL) 
		append_copy_symbols = cs;
	else {
		tmp=append_copy_symbols;
		while (tmp->next) 
			tmp=tmp->next;
		tmp->next = cs;
	}
        if (globalEnvPtr->debugFlag != 0 ) {
           for (tmp = append_copy_symbols, i=0; tmp!=NULL; tmp=tmp->next, i++) {
        	fprintf(stderr, "lextrace: add_copy_replacement dump4(%d): value=%s replacement=%s\n",
                         i,
                         tmp->value,
                         tmp->replacement 
                         );
           }
        }
}

void copy_do_replacements(char *buffer) {
	struct copy_symbols *cs;
	char *stripped, *s, *s1, *s2, *tmp, c;
	int len, end, i, sw, cnt;
	
	if (globalEnvPtr->debugFlag != 0 ) {
              fprintf(stderr, "lextrace: copy_do_replacements fsp = %d:\n",
    	            fsp
    	            );
        }
	// Replacement table empty
        copy_symbols = file_stack[fsp].last_symbol;
	if (file_stack[fsp].last_symbol == NULL) {
	   if (globalEnvPtr->debugFlag != 0 ) {
               fprintf(stderr, "lextrace: Empty replace list \n");
           }
	   return;
	}
        else {
	   if (globalEnvPtr->debugFlag != 0 ) {
             fprintf(stderr, "lextrace: Replace list is NOT empty \n");
             for (cs=copy_symbols, i=0; cs != NULL; cs=cs->next, i++) {
                  fprintf(stderr, "lextrace: Replace list (%02d) :%s: by :%s:\n",
                          i,
			 cs->value,
 			 cs->replacement
    	          	  );
             }
           }
        }

	// ignore comments
	if (globalEnvPtr->codeFormat == 1) {
	   if (*buffer == CHR_ASTERIX)  {
	      if (globalEnvPtr->debugFlag != 0 ) {
                 fprintf(stderr, "lextrace: Ignoring comment line :%s:\n", buffer);
              }
	      return;
	   }
	}
	else {
	   len = strlen(buffer);
	   if (len > FIXFORM_7) {
	      if (*(buffer + FIXFORM_6) == CHR_ASTERIX)  {
	         if (globalEnvPtr->debugFlag != 0 ) {
                    fprintf(stderr, "lextrace: Ignoring comment line :%s:\n", buffer);
                 }
	         return;
	      }
	   }
	   else {
	      if (globalEnvPtr->debugFlag != 0) {
                 fprintf(stderr, "lextrace: Ignoring empty fixed format line :%s:\n", buffer);
              }
	      return;
	   }
	}

	if (globalEnvPtr->debugFlag != 0 ) {
           fprintf(stderr, "lextrace: After comment check \n");
           fprintf(stderr, "lextrace: Before copy replace :%s:\n", buffer);
        }
	
	tmp = malloc(MAXLINEBUF);
 	strcpy(tmp, "");
	s  = buffer;
	s1 = tmp;
 	len = strlen(buffer);
	if (globalEnvPtr->codeFormat == 0) {
	   for (i=0; i<FIXFORM_7; i++) {
	      *s1++ = *s++;
	   }
           *(s1+1) = '\0';
	   i = FIXFORM_7;
	   if (len > FIXFORM_72) {
	      end = len - (len - FIXFORM_72);
	   }
	   else {
	      end = len;
	   }
	}
	else {
	   i = 0;
	   end = len;
	}

        copy_symbols = file_stack[fsp].last_symbol;

        cnt = 0;
        for (i; i<end; ) {
  	    if (*s && (*s == ' ' || *s == '\t')) {
               *s1 = *s;
               s1++;
               *s1 = '\0';
               s++;
               i++;
  	    }
  	    else {
  	       sw = 0;
               for (cs=copy_symbols; cs != NULL; ) {
                   if (!strncmp(s, cs->value, strlen(cs->value))) {
// 				c = *(s+strlen(cs->value));
// 				if (c==' ' || c =='\0') {
                                    strcat(s1, cs->replacement);
                                    cnt++;
				    s  += strlen(cs->value);
				    i  += strlen(cs->value); 
                                    s1 += strlen(cs->replacement);
//                                    *s1++ = ' ';
//                                    *(s1+1) = '\0';
				    if (globalEnvPtr->debugFlag != 0 ) {
                                       fprintf(stderr, "lextrace: Now replacing :%s: by :%s:\n",
    	                                    cs->value,
    	                                    cs->replacement
    	                                    );
                                    }
		                    cs = NULL;
  	                            sw = 1;
		   }
		   else {
		      cs=cs->next;
		   }
               }
               if (sw == 0) {
                  *s1 = *s;
                  s++;
                  s1++;
                  *s1 = '\0';
                  i++;
               }
            }
        }

        if (cnt != 0) {
           // Try to fix 73-80 cols.         
           if (globalEnvPtr->codeFormat == 0) {
           
              len = strlen(tmp);
	      if (globalEnvPtr->debugFlag != 0 ) {
                  fprintf(stderr, "lextrace: Before 73-80 cols fix len=%d :%s:\n",
    	              len,
    	              tmp
    	              );
              }
                 if (len > FIXFORM_72) {
                     sw = 0;
                     for (i=len-1; i>=FIXFORM_71; i--) {
                     	  s1 = tmp + i;
                     	  if (*s1 != ' ') {
                     	     sw++;
                     	     i = 0;
                     	  }
                     	  else {
                     	     s1 = tmp + i;
                     	     *s1 = '\0';
                     	  }
                     }
                     s1 = buffer + FIXFORM_71;
                     strcat(tmp, s1);
                 }
                 else {
                      if (len < FIXFORM_72) {
                         for (i=len - 1; i<FIXFORM_72; i++) {
                              s1 = tmp + i;
                              *s1 = ' ';
                              *(s1+1) = '\0';
                         }
                      }
                      s1 = buffer + FIXFORM_72;
                      strcat(tmp, s1);
                 }
           }
           strcpy(buffer, tmp);
        }
        
        if (globalEnvPtr->debugFlag != 0 ) {
           fprintf(stderr, "lextrace: After copy replace loop :%s:", buffer);
        }
	free(tmp);
}

/* 
  Despite the name, copy_strip_spaces eliminates ", " and "; "
  separators too, besides multiples spaces.
  This follows standard CD-1.2 pp.679, rule 7.b 
*/
char *copy_strip_spaces(char *s) {
	char *s1,*new;
	s1 = new = malloc(strlen(s)+2);
	while (*s==' ' || *s=='\t') /* don't strip indenting */
		*s1++=*s++;
	while (*s) {
		if ((*s==',' || *s==';') && *(s+1)==' ') s++;
		if (*s==' ' || *s=='\t' || *s=='\n') {
			*s1++ = ' ';
			if (*s=='\n') break;
			while (*s && (*s==' ' || *s=='\t')) s++;
		}
		else
			*s1++ = *s++;
	}
	*s1=0;
	return new;
}

